
package org.jibx.xsd2jibx;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Command line interface for the Xsd2Jibx code generation.
 */
public class Generate
{
    // base directory for code generation
    private String m_generateDirectory = ".";
    
    // default package used for no-namespace schemas
    private String m_defaultPackage = "";
    
    // substitution match and replacement strings
    private ArrayList m_pathLeadMatch;
    private ArrayList m_pathLeadReplace;
    
    private Generate() {
        m_pathLeadMatch = new ArrayList();
        m_pathLeadReplace = new ArrayList();
    }
    
    /**
     * Convert a namespace URI to a package path. 
     * 
     * @param uri namespace URI to be converted
     */
    private String pathForNamespace(String uri) {
        
        // first check for match to a replacement specification
        String path = uri;
        for (int i = 0; i < m_pathLeadMatch.size(); i++) {
            String match = (String)m_pathLeadMatch.get(i);
            if (path.startsWith(match)) {
                return (String)m_pathLeadReplace.get(i) +
                    path.substring(match.length()).toLowerCase();
            }
        }
        
        // if no replacement match, just use generic handling
        int split = path.indexOf(':');
        if (split >= 0) {
            path = path.substring(split+1);
            if (path.startsWith("//")) {
                
                // HTTP-type URI, turn into rational package name
                path = path.substring(2);
                if (path.startsWith("www.")) {
                    path = path.substring(4);
                }
                
                // split trailing portion following path separator
                String trailing = "";
                split = path.indexOf('/');
                if (split >= 0) {
                    trailing = path.substring(split+1);
                    path = path.substring(0, split);
                }
                
                // reverse the order of components of domain name
                StringBuffer reversed = new StringBuffer();
                while ((split = path.lastIndexOf('.')) >= 0) {
                    reversed.append('.');
                    reversed.append(path.substring(split+1));
                    path = path.substring(0, split);
                }
                reversed.append('.');
                reversed.append(path);
                
                // process path components following domain name
                path = reversed.substring(1);
                while (trailing.length() > 0) {
                    split = trailing.indexOf('/');
                    String part;
                    if (split >= 0) {
                        part = trailing.substring(0, split);
                        trailing = trailing.substring(split+1);
                    } else {
                        part = trailing;
                        trailing = "";
                    }
                    part = part.replace(':', '_').replace('-', '_').
                        replace('.', '_');
                    if (Character.isDigit(part.charAt(0))) {
                        part = '_' + part;
                    }
                    path += "/" + part;
                }
            }
        }
        path = path.toLowerCase().replace('/', '.');
        int last = path.length() - 1;
        if (path.charAt(last) == '.') {
            path = path.substring(0, last);
        }
        return path;
    }
    
    /**
     * Generate JiBX code and bindings for a list of schema definitions.
     * 
     * @param schemas list of schema elements to be processed
     */
    private void generate(List schemas) {
        
        // set up for generation of all schemas
        Generator gen = new Generator();
        GeneratorConfig cfg = new GeneratorConfig();
        cfg.setOutputDir(m_generateDirectory);
        cfg.setBindingFile("binding.xml");
        cfg.setTargetPackage(m_defaultPackage);
        cfg.setXmlNamespaceToJavaPackageMap(new PackageMap());
        
        // add all schemas to be processed
        for (Iterator iter = schemas.iterator(); iter.hasNext();) {
            String path = (String)iter.next();
            FileXsdResolver resolver = new FileXsdResolver(path);
            cfg.addXsdResolver(resolver);
        }
        
        // handle the actual generation
        gen.setConfig(cfg);
        gen.execute();
        gen.writeJavaSources();
        gen.marshalJibxBindingDefinition();
    }
    
    /**
     * Command line interface for Xsd2Jibx generation.
     * 
     * @param args command line arguments
     */
    public static void main(String[] args) {
        
        // make sure we have schema to process
        boolean help = false;
        if (args.length > 0) {
            
            // process leading special arguments
            Generate inst = new Generate();
            int offset = 0;
            ArrayList remaps = new ArrayList();
            for (; offset < args.length-1; offset++) {
                String arg = args[offset];
                if ("-d".equalsIgnoreCase(arg)) {
                    inst.m_generateDirectory = args[++offset];
                } else if ("-p".equalsIgnoreCase(arg)) {
                    String spec = args[++offset];
                    int split = spec.indexOf('=');
                    if (split > 0) {
                        inst.m_pathLeadMatch.add(spec.substring(0, split));
                        inst.m_pathLeadReplace.add(spec.substring(split+1));
                    } else {
                        System.err.println("Invalid URI substitution: " + spec);
                        System.exit(1);
                    }
                } else if ("-n".equalsIgnoreCase(arg)) {
                    inst.m_defaultPackage = args[++offset];
                } else {
                    break;
                }
            }
            if (offset < args.length && args[offset].startsWith("-")) {
                System.err.println("Unknown argument " + args[offset]);
                help = true;
            } else if (offset < args.length) {
                
                // process all remaining arguments as schemas
                ArrayList schemas = new ArrayList();
                while (offset < args.length) {
                    schemas.add(args[offset++]);
                }
                try {
                    inst.generate(schemas);
                } catch (Throwable t) {
                    t.printStackTrace();
                    System.exit(1);
                }
                
            } else {
                System.err.println("Need one or more schemas to process");
                help = true;
            }
        }
        
        // print usage information if problem with parameters
        if (help) {
            System.out.println
                ("\nUsage: java org.jibx.xsd2jibx.Generate " +
                "[-options] schema1 schema2 ...\nWhere options are:\n" +
                " -d  target directory for generated classes,\n" +
                " -n  package used for no-namespace schemas,\n" +
                " -p  path substitution for mapping namespaces to packages,\n" +
                "and the schemaN parameters are paths to one or more schemas " +
                "to be processed");
            System.exit(1);
        }
    }
    
    /**
     * Faked map for converting schema namespaces to packages. This is a very
     * partial and incomplete implementation that basically just fakes the
     * portions of the Map interface necessary for the conversion. Ugly, but
     * required by the design of the Xsd2Jibx API.
     */
    private class PackageMap implements Map
    {
        public int size() {
            return 0;
        }
        public void clear() {}
        public boolean isEmpty() {
            return false;
        }
        public boolean containsKey(Object key) {
            return true;
        }
        public boolean containsValue(Object value) {
            return false;
        }
        public Collection values() {
            return null;
        }
        public void putAll(Map t) {}
        public Set entrySet() {
            return null;
        }
        public Set keySet() {
            return null;
        }
        public Object remove(Object key) {
            return null;
        }
        public Object put(Object key, Object value) {
            return null;
        }
        /**
         * Get value for key. This is the only method that's actually
         * semi-implemented, since it's the only one actually used by the code.
         */
        public Object get(Object key) {
            return pathForNamespace((String)key);
        }
    }
}